/*
 * 3D Engine 
 * display.c: 
 * Based on pikuma.com 3D software renderer in C
 * Copyright (c) 2021 986-Studio. All rights reserved.
 *
 * Created by Manoël Trapier on 01/03/2021.
 */

#include <stdio.h>
#include <stdbool.h>

#include <SDL.h>

#include <log.h>
#include <display.h>

/***********************************************************************************************************************
 * Global variables
 **********************************************************************************************************************/

SDL_Window *window = NULL;
SDL_Renderer *renderer = NULL;
uint32_t *frameBuffer = NULL;
SDL_Texture *frameBufferTexture = NULL;
uint32_t windowWidth = 800;
uint32_t windowHeight = 600;

/***********************************************************************************************************************
 * Functions
 **********************************************************************************************************************/
#define MAX(_a, _b) ((_a) < (_b)) ? (_b) : (_a)
#define MIN(_a, _b) ((_a) > (_b)) ? (_b) : (_a)

bool initialiseWindow(bool fullScreen)
{
    bool ret = false;
    int windowFlags = 0;

    if (SDL_Init(SDL_INIT_EVERYTHING))
    {
        log(TLOG_PANIC, NULL, "SDL Initialisation error!! error: %s", SDL_GetError());
        goto exit;
    }

    log(TLOG_DEBUG, NULL, "SDL properly initialised!");

    if (fullScreen)
    {
        log(TLOG_DEBUG, NULL, "Will go fullscreen! Fasten your seatbelts!");
        SDL_DisplayMode displayMode;
        SDL_GetCurrentDisplayMode(0, &displayMode);

        windowWidth = displayMode.w;
        windowHeight = displayMode.h;

        windowFlags = SDL_WINDOW_BORDERLESS;
    }

    window = SDL_CreateWindow(NULL, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, windowWidth, windowHeight,
                              windowFlags);
    if (window == NULL)
    {
        log(TLOG_PANIC, NULL, "SDL Window creation error: %s", SDL_GetError());
        goto exit;
    }
    log(TLOG_DEBUG, NULL, "SDL Window created!");

    renderer = SDL_CreateRenderer(window, -1, 0);
    if (renderer == NULL)
    {
        log(TLOG_PANIC, NULL, "SDL Renderer creation error: %s", SDL_GetError());
        goto exit;
    }
    log(TLOG_DEBUG, NULL, "SDL Renderer created!");

    if (fullScreen)
    {
        SDL_SetWindowFullscreen(window, SDL_WINDOW_FULLSCREEN);
    }

    ret = true;

exit:
    return ret;
}

void destroyWindow()
{
    if (frameBuffer != NULL)
    {
        free(frameBuffer);
    }
    SDL_DestroyRenderer(renderer);
    SDL_DestroyWindow(window);
    SDL_Quit();
}

void renderFrameBuffer()
{
    SDL_UpdateTexture(frameBufferTexture, NULL, frameBuffer, (uint32_t)(windowWidth * sizeof(uint32_t)));
    SDL_RenderCopy(renderer, frameBufferTexture, NULL, NULL);
}

void drawPixel(uint32_t x, uint32_t y, uint32_t colour)
{
    if ((x >= 0) && (x < windowWidth) && (y >= 0) && (y < windowHeight))
    {
        frameBuffer[x + (y * windowWidth)] = colour;
    }
}

void drawVLine(uint32_t x, uint32_t y0, uint32_t y1, uint32_t colour)
{
    uint32_t i;
    if (y0 > y1) { i = y1; y1 = y0; y0 = i; }

    if ((x < 0) || (x > windowWidth))
    {
        return;
    }

    y0 = MAX(0, y0);
    y1 = MIN(y1, (windowHeight - 1));

    uint32_t *currentPos = &frameBuffer[x + (y0 * windowWidth)];

    for(i = y0; i < y1; i++, currentPos+=windowWidth)
    {
        *currentPos = colour;
    }
}

void drawHLine(uint32_t x0, uint32_t y, uint32_t x1, uint32_t colour)
{
    uint32_t i;
    if (x0 > x1) { i = x1; x1 = x0; x0 = i; }

    if ((y < 0) || (y > windowHeight))
    {
        return;
    }

    x0 = MAX(0, x0);
    x1 = MIN(x1, (windowWidth - 1));

    uint32_t *currentPos = &frameBuffer[x0 + (y * windowWidth)];

    for(i = x0; i < x1; i++, currentPos++)
    {
        *currentPos = colour;
    }
}

void clearFrameBuffer(uint32_t colour)
{
    uint32_t x, y;
    for (y = 0 ; y < windowHeight ; y++)
    {
        for (x = 0 ; x < windowWidth ; x++)
        {
            frameBuffer[x + (y * windowWidth)] = colour;
        }
    }
}

void drawGrid(int spacing, uint32_t colour)
{
    uint32_t x, y;
    for (y = 0 ; y < windowHeight ; y++)
    {
        for (x = 0 ; x < windowWidth ; x++)
        {
            if (((x % spacing) == 0) || ((y % spacing) == 0))
            {
                frameBuffer[x + (y * windowWidth)] = colour;
            }
        }
    }
}

void drawRectangle(uint32_t x, uint32_t y, uint32_t w, uint32_t h, uint32_t colour)
{
    uint32_t j;

    for (j = 0 ; j < h ; j++)
    {
        drawHLine(x, y + j, x + w, colour);
    }
}

/* FIXME: Some out of screen line may render weirdly. Will fix that later */
void drawLine(uint32_t x0, uint32_t y0, uint32_t x1, uint32_t y1, uint32_t colour)
{
    int i;
    int32_t deltaX = (int32_t)x1 - (int32_t)x0;
    int32_t deltaY = (int32_t)y1 - (int32_t)y0;

    int32_t sideLength = ( abs(deltaX) >= abs(deltaY) ) ? abs(deltaX) : abs(deltaY);

    double incrementX = deltaX / (double)sideLength;
    double incrementY = deltaY / (double)sideLength;

    if (deltaX == 0)
    {
        drawVLine(x0, y0, y1, colour);
    }
    else if (deltaY == 0)
    {
        drawHLine(x0, y0, x1, colour);
    }
    else
    {
        double currentX = (int32_t)x0;
        double currentY = (int32_t)y0;

        for (i = 0 ; i < sideLength ; i++)
        {
            drawPixel(round(currentX), round(currentY), colour);
            currentX += incrementX;
            currentY += incrementY;
        }
    }
}

void drawTriangle(uint32_t x0, uint32_t y0, uint32_t x1, uint32_t y1, uint32_t x2, uint32_t y2, uint32_t colour)
{
    drawLine(x0, y0, x1, y1, colour);
    drawLine(x1, y1, x2, y2, colour);
    drawLine(x2, y2, x0, y0, colour);
}